@@ -29,7 +29,7 @@ from apostrophe import helpers
 from apostrophe.editor import Editor
 from apostrophe.export_dialog import AdvancedExportDialog, ExportDialog
 from apostrophe.headerbars import BaseHeaderbar
-from apostrophe.helpers import App
+from apostrophe.helpers import App, bind_enum
 from apostrophe.panels import ApostrophePanels
 from apostrophe.preview_handler import PreviewHandler
 from apostrophe.preview_security import PreviewSecurity
@@ -40,6 +40,7 @@ from apostrophe.text_view import ApostropheTextView
 from apostrophe.text_view_format_inserter import FormatInserter
 from apostrophe.preview_security import PreviewSecurityHandler
 from apostrophe.pride import apply_seasonal_style
+from apostrophe.sized_bin import ApostropheSizedBin
 
 LOGGER = logging.getLogger('apostrophe')
 
@@ -52,8 +53,9 @@ class MainWindow(Adw.ApplicationWindow):
 
     editor = Gtk.Template.Child()
     save_progressbar = Gtk.Template.Child()
-    toolbar_revealer = Gtk.Template.Child()
+    headerbar_revealer = Gtk.Template.Child()
     headerbar = Gtk.Template.Child()
+    headerbar_narrow = Gtk.Template.Child()
     searchbar = Gtk.Template.Child()
     panels = Gtk.Template.Child()
     preview_stack = Gtk.Template.Child()
@@ -62,6 +64,9 @@ class MainWindow(Adw.ApplicationWindow):
     preview_spinner = Gtk.Template.Child()
     webview_snapshot = Gtk.Template.Child()
 
+    headerbars_breakpointbin = Gtk.Template.Child()
+    headerbars_breakpoint = Gtk.Template.Child()
+
     subtitle = GObject.Property(type=str)
     is_fullscreen = GObject.Property(type=bool, default=False)
 
@@ -70,6 +75,9 @@ class MainWindow(Adw.ApplicationWindow):
 
     did_change = GObject.Property(type=bool, default=False)
     current = GObject.Property(type=GObject.Object, default=None)
+    snapshot = GObject.Property(type=Gio.File, default=None)
+    snapshot_restored = GObject.Property(type=bool, default=False)
+    topbars_height = GObject.Property(type=int, default=0)
 
     close_anyway = False
     file_monitor = None
@@ -80,21 +88,23 @@ class MainWindow(Adw.ApplicationWindow):
         super().__init__(application=Gio.Application.get_default(),
                          title="Apostrophe")
 
-        #TODO: size
-
         # Preferences
         self.settings = Settings.new()
 
         # Connect signals that we can't connect on the UI file
         self.connect("notify::is-fullscreen", self._on_fullscreen)
+
         # Create new, empty file
         # TODO: load last opened file?
-
         self.current = File()
 
         # Setup text editor
         self.textview = self.editor.textview
-        self.textview.get_buffer().connect('changed', self.on_text_changed)
+        self.text_changed_handler_id = None
+        self.autosave_timer = None
+
+        # save metadata for spellchecking language
+        self.textview.spelling_checker.connect("notify::language", self._on_spellchecking_language_changed)
 
         # Setup save progressbar an its animator
         def hide_progressbar(animation, *args):
@@ -121,9 +131,6 @@ class MainWindow(Adw.ApplicationWindow):
         apply_seasonal_style(self.save_progressbar)
 
         # Setup preview
-
-        self.preview_layout = self.settings.get_enum("preview-mode")
-
         self.preview_handler = PreviewHandler(self, self.textview, self.panels)
 
         # not really necessary but we'll keep a preview_layout property on the window
@@ -139,16 +146,11 @@ class MainWindow(Adw.ApplicationWindow):
         # preview security handler
         self.preview_security_handler = PreviewSecurityHandler(self)
 
-        # Setting up spellcheck
-        #self.settings.bind("spellcheck", self.textview,
-        #                   "spellcheck", Gio.SettingsBindFlags.GET)
-
-        # Setting up text size
-        self.settings.bind("bigger-text", self.textview,
-                           "bigger_text", Gio.SettingsBindFlags.GET)
-
-        self.settings.bind("characters-per-line", self.textview,
-                           "line_chars", Gio.SettingsBindFlags.GET)
+        # Headerbars breakpoint bin minimum size
+        self.headerbars_breakpointbin.set_size_request(
+            self.headerbar_narrow.measure(Gtk.Orientation.HORIZONTAL, -1).minimum,
+            self.headerbar_narrow.measure(Gtk.Orientation.VERTICAL, -1).minimum
+            )
 
         # Search and replace initialization
         self.searchbar.attach(self.textview)
@@ -277,20 +279,31 @@ class MainWindow(Adw.ApplicationWindow):
         action.connect_after("activate", FormatInserter().insert_table, self.textview)
         self.add_action(action)
 
-        scrollbar = self.editor.scrolledwindow.get_vscrollbar()
-        scrollbar.set_margin_top(54)
-        scrollbar.set_margin_bottom(48)
+        # update textview's top margin on topbar size changes
+        self.connect("notify::topbars-height", self.update_textview_margin)
 
-
-        self.preview = self.settings.get_boolean("preview-active")
-
-        self.textview.buffer.hemingway_mode = self.settings.get_boolean("hemingway-mode")
+        # Bind gsettings
+        self.settings.bind("window-width", self, "default-width",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
+        self.settings.bind("window-height", self, "default-height",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
+        self.settings.bind("is-maximized", self, "maximized",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
+        self.settings.bind("is-fullscreen", self, "fullscreened",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
+        bind_enum(self.settings, "preview-mode", self, "preview-layout",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
+        self.settings.bind("preview-active", self, "preview",
+            Gio.SettingsBindFlags.DEFAULT|Gio.SettingsBindFlags.GET_NO_CHANGES)
 
         self.new_document()
 
         if self.get_application()._application_id == 'org.gnome.gitlab.somas.Apostrophe.Devel':
             self.add_css_class('devel')
 
+    def update_textview_margin(self, *args, **kwargs):
+        self.textview.update_vertical_margin(self.topbars_height)
+        self.textview.queue_resize()
 
     def on_text_changed(self, *_args):
         """called when the text changes, sets the self.did_change to true and
@@ -299,6 +312,13 @@ class MainWindow(Adw.ApplicationWindow):
 
         if self.did_change is False:
             self.did_change = True
+
+            # Autosaving
+            self.autosave_timer = GLib.timeout_add_seconds(self.settings.get_int("autosave-period"), self.autosave_snapshot)
+
+        if self.snapshot_restored is True:
+            self.snapshot_restored = False
+
         self.update_headerbar_title(True, True)
         if self.settings.get_value("autohide-headerbar"):
             self.hide_headerbar_bottombar()
@@ -380,6 +400,10 @@ class MainWindow(Adw.ApplicationWindow):
                     cancellable=None,
                     callback=self._replace_contents_cb,
                     user_data=callback)
+                
+                if self.snapshot_restored:
+                    self.snapshot_restored = False
+
         # if there's no GFile we ask for one:
         else:
             self.save_document_as(callback=callback)
@@ -460,6 +484,10 @@ class MainWindow(Adw.ApplicationWindow):
             recents_manager.add_item(self.current.gfile.get_uri())
             if callback is not None:
                 callback(self)
+
+            # remove the snapshot as everything is secure on disc.
+            # it'll be recreated if the document changes
+            self.delete_snapshot()
         else:
             self.progressbar_fade_out.play()
             self.did_change = True
@@ -522,6 +550,12 @@ class MainWindow(Adw.ApplicationWindow):
         self.current.gfile = file
         self.preview_security_handler.set_file_security_level()
 
+        if self.text_changed_handler_id:
+            self.textview.get_buffer().disconnect(self.text_changed_handler_id)
+
+        if spelling_language := file.query_info("metadata::spelling-language", Gio.FileQueryInfoFlags.NONE).get_attribute_string("metadata::spelling-language"):
+            self.textview.spelling_checker.set_language(spelling_language)
+
         self.current.gfile.load_contents_async(None,
                                                self._load_contents_cb, None)
         self._set_file_monitor()
@@ -544,7 +578,12 @@ class MainWindow(Adw.ApplicationWindow):
     def _on_file_changed(self, file, other_file, event, *args):
         self.discard_infobar.set_revealed(True)
 
-    def _load_contents_cb(self, gfile, result, user_data=None):
+    def _on_spellchecking_language_changed(self, *args, **kwargs):
+        if language:= self.textview.spelling_checker.get_language():
+            if self.current.gfile:
+                self.current.gfile.set_attribute_string("metadata::spelling-language", language, Gio.FileQueryInfoFlags.NONE)
+
+    def _load_contents_cb(self, gfile, result, snapshot_restored=False):
         try:
             _success, contents, _etag = gfile.load_contents_finish(result)
         except GLib.GError as error:
@@ -574,20 +613,30 @@ class MainWindow(Adw.ApplicationWindow):
             return
         else:
             self.textview.set_text(decoded)
+
             self.preview_security_handler.assert_security_risk(decoded)
             start_iter = self.textview.get_buffer().get_start_iter()
             GLib.idle_add(
                 lambda: self.textview.get_buffer().place_cursor(start_iter))
 
-            ## add file to recents manager once it's fully loaded,
+            # reconnect the text changed handler
+            self.text_changed_handler_id = self.textview.get_buffer().connect('changed', self.on_text_changed)
+
+            # add file to recents manager once it's fully loaded,
             # unless it is an internal resource
-            if not self.current.gfile.get_uri().startswith("resource:"):
+            if self.current.gfile and not self.current.gfile.get_uri().startswith("resource:"):
                 recents_manager = Gtk.RecentManager.get_default()
                 recents_manager.add_item(self.current.gfile.get_uri())
 
             self.update_headerbar_title()
 
-            self.did_change = False
+            # when restoring a snapshot we want to preserve the "did_change" state
+            # because we restored a dirty state, but we need to revert the snapshot_restored
+            # back and show again the headerbar
+            if snapshot_restored:
+                self.textview.get_buffer().emit("changed")
+                self.snapshot_restored = True
+                self.reveal_headerbar_bottombar()
 
     def check_change(self,
                      callback = None):
@@ -631,6 +680,7 @@ class MainWindow(Adw.ApplicationWindow):
         """
         def callback(self):
             self.textview.clear()
+            self.text_changed_handler_id = self.textview.get_buffer().connect('changed', self.on_text_changed)
 
             self.did_change = False
             self.current.gfile = None
@@ -680,17 +730,21 @@ class MainWindow(Adw.ApplicationWindow):
     def reveal_headerbar_bottombar(self, *args):
         self.editor.reveal_bottombar()
 
-        if not self.toolbar_revealer.get_reveal_child():
-            self.toolbar_revealer.set_reveal_child(True)
+        if not self.headerbar_revealer.get_reveal_child():
+            self.headerbar_revealer.set_reveal_child(True)
             self.remove_css_class("no-headerbar")
 
     def hide_headerbar_bottombar(self):
-        if self.searchbar.search_mode_enabled or\
-           self.discard_infobar.get_revealed():
+        if self.headerbars_breakpointbin.get_current_breakpoint() == self.headerbars_breakpoint:
             return
 
-        if self.toolbar_revealer.get_reveal_child():
-            self.toolbar_revealer.set_reveal_child(False)
+        if self.searchbar.search_mode_enabled or\
+           self.discard_infobar.get_revealed() or\
+           self.snapshot_restored:
+            return
+
+        if self.headerbar_revealer.get_reveal_child():
+            self.headerbar_revealer.set_reveal_child(False)
             self.add_css_class("no-headerbar")
 
         self.editor.hide_bottombar()
@@ -721,13 +775,6 @@ class MainWindow(Adw.ApplicationWindow):
         self.subtitle = subtitle
         self.headerbar.set_tooltip_text(subtitle)
 
-    def save_state(self):
-        self.settings.set_enum("preview-mode", self.preview_layout)
-        self.settings.set_boolean("preview-active", self.preview)
-        self.settings.set_boolean("hemingway-mode", self.textview.buffer.hemingway_mode)
-        self.settings.set_boolean("toolbar-active", self.editor.toolbar_revealer.extra_toolbar_revealed)
-        self.settings.set_boolean("spellcheck", self.textview.adapter.get_enabled())
-
     def do_close_request(self, *args):
         LOGGER.info('close request called')
 
@@ -737,24 +784,77 @@ class MainWindow(Adw.ApplicationWindow):
 
         # called if check_change decides we can throw away the contents of the textview
         def callback(window):
-            # save state if we're the last window group left
-            n_windows = len(window.get_application().get_main_windows())
-
-            if n_windows == 1:
-                window.save_state()
             window.close_anyway = True
             window.do_close_request()
 
         self.check_change(callback)
         return True
 
-    @Gtk.Template.Callback()
-    def spinner_map_cb(self, *args):
-        self.preview_spinner.set_spinning(True)
+    def autosave_snapshot(self):
+        try:
+            try:
+                encoded_text = self.textview.get_text()\
+                    .encode(self.current.encoding)
+            except UnicodeEncodeError:
+                encoded_text = self.textview.get_text()\
+                    .encode("UTF-8")
+                self.current.encoding = "UTF-8"
+        except UnicodeEncodeError as error:
+            LOGGER.warning(str(error.reason))
+            return
+        if not encoded_text:
+            return
+        else:
+            if not self.snapshot:
+                import uuid
+                snapshot_dir = GLib.build_filenamev([GLib.get_user_state_dir(), "snapshots"])
+                GLib.mkdir_with_parents(snapshot_dir, 0o750)
+                snapshot_path = GLib.build_filenamev([snapshot_dir, uuid.uuid4().hex])
+                self.snapshot = Gio.File.new_for_path(snapshot_path)
+            self.snapshot.replace_contents_bytes_async(
+                GLib.Bytes.new(encoded_text),
+                etag=None,
+                make_backup=False,
+                flags=Gio.FileCreateFlags.NONE,
+                cancellable=None,
+                callback=self._autosave_finish)
+            return True
 
-    @Gtk.Template.Callback()
-    def spinner_unmap_cb(self, *args):
-        self.preview_spinner.set_spinning(False)
+    def _autosave_finish(self, gfile, result, callback=None):
+        try:
+            success, _etag = gfile.replace_contents_finish(result)
+        except GLib.GError as error:
+            LOGGER.warning(str(error.message))
+            return
+        if success and self.current.gfile:
+            self.snapshot.set_attribute_string("metadata::original-path", self.current.gfile.get_path(), Gio.FileQueryInfoFlags.NONE)
+        return success
+    
+    def load_snapshot(self, file):
+        """Load a snapshot file contents while setting apostrophe's gfile to the actual user file."""
+        LOGGER.info("trying to load snapshot %s", file.get_uri())
+
+        self.snapshot = file
+        if gfile_path := file.query_info("metadata::original-path", Gio.FileQueryInfoFlags.NONE).get_attribute_string("metadata::original-path"):
+            self.current.gfile = Gio.File.new_for_path(gfile_path)
+
+        self.snapshot_restored = True
+        self.textview.get_buffer().disconnect(self.text_changed_handler_id)
+        self.snapshot.load_contents_async(None, self._load_contents_cb, self.snapshot_restored)
+        self._set_file_monitor()
+
+    def delete_snapshot(self):
+        # delete snapshots in the application in case the deletion outlives the window
+        if self.autosave_timer:
+            GLib.Source.remove(self.autosave_timer)
+            self.autosave_timer = None
+        if self.snapshot:
+            self.get_application().delete_snapshot(self.snapshot)
+            self.snapshot = None
+
+    def do_dispose(self):
+        self.delete_snapshot()
+        super().do_dispose()
 
 #@dataclass
 class File(GObject.Object):
@@ -804,3 +904,7 @@ class File(GObject.Object):
         self.title = _("New File")
         self.name = ""
         self.security_level = PreviewSecurity.ASK
+
+    def do_dispose(self):
+        self.gfile = None
+        super().do_dispose()
