@@ -13,7 +13,7 @@
 # with this program.  If not, see <http://www.gnu.org/licenses/>.
 # END LICENSE
 
-import re
+import regex as re
 from multiprocessing import Pipe, Process
 
 import gi
@@ -210,7 +210,7 @@ class MarkupHandler:
                 code_blocks.append((start, end))
 
             # helper functions to test if a range is inside code blocks.
-            inside_code_blocks = lambda start, end: any(True for (s, e) in code_blocks if s < start and end <  e)
+            inside_code_blocks = lambda start, end: any(True for (s, e) in code_blocks if s < start or end <  e)
             match_inside_code_blocks = lambda m: inside_code_blocks(m.start(), m.end())
 
             # Find:
@@ -301,6 +301,10 @@ class MarkupHandler:
             for match in matches:
                 if match_inside_code_blocks(match):
                     continue
+                # check if it's a YAML fronmatter
+                if frontmatter := re.search(markup_regex.FRONTMATTER, text):
+                    if match.start() <= frontmatter.end():
+                        continue
                 result.append(
                     (self.TAG_NAME_BOLD, (), match.start(), match.end()))
 
